//
//! \file  libecc/rds.h
//! \brief Random Digits Stream.
//!
//! This header file declares <code>class libecc::rds</code>, a generator of pseudo random digits.
//
// This file is part of the libecc package.
// Copyright (C) 2004, by
//
// Carlo Wood, Run on IRC <carlo@alinoe.com>
// RSA-1024 0x624ACAD5 1997-01-26                    Sign & Encrypt
// Fingerprint16 = 32 EC A7 B6 AC DB 65 A6  F6 F6 55 DD 1C DC FF 61
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#ifndef LIBECC_RDS_H
#define LIBECC_RDS_H

#include <libecc/rng.h>
#include <libecc/sha.h>

namespace libecc {

class rds {
private:
  rng M_rng;
  unsigned int M_available;
  bitset_digit_t const* M_bufptr;
  sha1 M_hash;
  bool M_use_SHA1;

public:
  /**
   * \brief The type of the constructor argument.
   */
  typedef rng::pool_type pool_type;

  /**
   * \brief Constructor.
   *
   * \param seed A bitset of 521 bits, any value is allowed except all zeroes.
   * \param use_SHA1 Use SHA1 on the output of the underlaying rng when set.
   *
   * When \a use_SHA1 is not set then the class is much faster, but not
   * cryptographically safe because the output can be predicted from
   * previous output.
   */
  rds(pool_type const& seed, bool use_SHA1 = true) : M_rng(seed), M_available(0), M_use_SHA1(use_SHA1) { }

  void read(bitset_digit_t* outptr, unsigned int digits);

  /**
   * \brief Fill bitset \a out with random bits.
   */
  template<unsigned int n>
    void read(bitset<n>& out)
    {
      bitset_digit_t* outptr = out.digits_ptr();
      this->read(outptr, bitset_base<n>::digits);
      if (bitset_base<n>::has_excess_bits)
        outptr[bitset_base<n>::digits - 1] &= bitset_base<n>::valid_bits; 
    }
};

} // namespace libecc

#endif // LIBECC_RDS_H
